/*
 *  Audit helper functions used throughout sudo
 *
 *  Copyright (C) 2007, Red Hat, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <config.h>

#ifdef WITH_AUDIT
#include <stdlib.h>
#include <syslog.h>
#include <stdarg.h>
#include <libaudit.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>

#ifdef HAVE_SELINUX
#include <selinux/selinux.h>
#endif

int audit_fd = -1;

void audit_help_open (void)
{
	audit_fd = audit_open ();
	if (audit_fd < 0) {
        	/* You get these only when the kernel doesn't have
                 * audit compiled in. */
		if (errno == EINVAL || errno == EPROTONOSUPPORT ||
		    errno == EAFNOSUPPORT)
			return;
		fprintf (stderr, "Cannot open audit interface - aborting.\n");
		exit (1);
	}
}

/*
 * This function will log a message to the audit system using a predefined
 * message format. Parameter usage is as follows:
 *
 * type - type of message: AUDIT_USER_CMD
 * command - the command being logged
 * params - parames of the command
 * result - 1 is "success" and 0 is "failed"
 *
 */
void audit_logger (int type, const char *command, const char *params, int result)
{
	int err;
	char *msg;

	if( audit_fd < 0 )
		return;
	else {

		if( params ) 
	                err = asprintf(&msg, "%s %s", command, params);
		else 
	                err = asprintf(&msg, "%s", command);
                if (err < 0) {
                        fprintf (stderr, "Memory allocation for audit message wasn’t possible.\n");
                        return;
                }
		
		err = audit_log_user_command (audit_fd, type, msg, NULL, result);
               /* The kernel supports auditing and we had 
                  enough privilege to write to the socket. */
		if( err <= 0 && !((errno == EPERM && getuid() > 0) || errno == ECONNREFUSED ) ) {
			perror("audit_log_user_command()");
		}

		free(msg);
	}
}

#ifdef HAVE_SELINUX
int send_audit_message(int success, security_context_t old_context,
                       security_context_t new_context, const char *ttyn)
{
        char *msg = NULL;
        int rc;

        if (audit_fd < 0)
                return -1;

        if (asprintf(&msg, "newrole: old-context=%s new-context=%s",
                     old_context ? old_context : "?",
                     new_context ? new_context : "?") < 0) {
                fprintf(stderr, "Error allocating memory.\n");
                rc = -1;
                goto out;
        }

        rc = audit_log_user_message(audit_fd, AUDIT_USER_ROLE_CHANGE,
                                    msg, NULL, NULL, ttyn, success);

        if (rc <= 0) {
                fprintf(stderr, "Error sending audit message.\n");
                rc = -1;
                goto out;
        }
        rc = 0;

        out:
         free(msg);
         return rc;
}
#endif
#endif                         /* WITH_AUDIT */
